#!/bin/bash
set -e #-x

# DO NOT ADD NEW EXPORTED VARIABLES HERE

# Add new operands in this list.
available_operands="custom rootfromvcs synconly nosync regendefconfig noconf nocleanstaging nobuild notrackws noforcesync noccache fast doc tests enable_tracing createpackage"

usage() {
  echo -n "-- Usage: $0 <project directory> " >&2
  for op in $available_operands; do
    echo -n "[$op] " >&2
  done
  echo ""
  exit 1
}
if [ -z "$1" ] || [ ! -d "$1" ]; then
  usage
fi

curdir=`pwd`
prjdef=$1
prjdir=`cd $prjdef;pwd`
starttime=`date +%Y-%m-%d_%H:%M:%S`

echo "-- Build started at: ${starttime}"
echo "-- Project directory: ${prjdir}"

prjdep="$prjdir/project.list"

hardcofile="$prjdir/hardco.cfg"

for f in "$prjdep"; do
  if ! [ -r "$f" ]; then
    n=$(basename "$f")
    echo "unable to read config file $n"
    echo "check that it exists in $prjdef"
    exit 1
  fi
done

# Set variables to "1" for the optional parameters. Name those variables to the parameters
shift 1
while [ -n "$1" ]; do
  invalid=true
  for op in $available_operands; do
    if [ "$op" = "$1" ]; then
  echo $1 is set
  export $1=1
      invalid=false
    fi
  done

  if $invalid; then
    echo "$1: invalid operand."
    exit 1
  fi

  shift 1
done

if [ -f server.cfg ]; then
  servercfgfile="server.cfg"
else
  if [ -f "$HOME/.server.cfg" ]; then
   servercfgfile="$HOME/.server.cfg"
  else
    if [ -f "$HOME/server.cfg" ]; then
      servercfgfile="$HOME/server.cfg"
    else
      echo "-- ERROR: server.cfg not found. This file should be in this directory or"
      echo "-- create a file .server.cfg or server.cfg in your home directory."
      echo "-- It should contain a list of perforce servers and your workspace configuration."
      echo "-- Check the sample directory for sample files."
      exit 1
    fi
  fi
fi
echo "-- Server configuration file: $servercfgfile"

if [ -n "${OUTPUTDIR}" ]; then
  if [ "`echo ${OUTPUTDIR} | cut -c1`" = "/" ]; then
    outputdir="${OUTPUTDIR}"
  else
    outputdir="`(cd "${OUTPUTDIR}"; pwd)`"
  fi
else
  outputdir="$curdir/output"
fi
echo "-- Output directory: $outputdir"

STAGINGDIR=$outputdir/staging
echo "-- Staging directory: $STAGINGDIR"

# Only auto-clean staging dir when noconf isn't set
if [ "${nocleanstaging}" != "1" ] && [ "${noconf}" != "1" ] && [ "${tests}" != "1" ]; then
  echo "-- Cleaning $STAGINGDIR"
  rm -rf $STAGINGDIR
fi

mkdir -p $STAGINGDIR

RELEASEDIR=$outputdir/release
BUILD_CUSTOMER=""
BUILD_PROJECT=""
BUILD_HARDWARE=""
BUILD_CONFIGTYPE=""
k=0
for i in `echo $prjdef |sed 's/\// /g'`; do
    arr[$k]=$i
    let k=$k+1
done
# n is total number of directories found
n=${#arr[*]}
BUILD_CUSTOMER=${arr[$n-4]}
BUILD_PROJECT=${arr[$n-3]}
BUILD_HARDWARE=${arr[$n-2]}
BUILD_CONFIGTYPE=${arr[$n-1]}

if [ "${nocleanstaging}" != "1" ]; then
  # remove the root filesystem directory
  echo "-- Cleaning $RELEASEDIR/root_fs"
  rm -rf $RELEASEDIR/root_fs
fi

echo "-- Release directory: $RELEASEDIR"
mkdir -p $RELEASEDIR

spinnumber=`grep CONFIG_BUILD_SPIN_NUMBER= $prjdir/defconfig | sed -e 's/CONFIG_BUILD_SPIN_NUMBER=//' | sed -e 's/\"//g' `
if [ ${#spinnumber} -eq 0 ]; then
	spinnumber="00"
fi

echo "-- Spin number: $spinnumber"

bldcfg="buildconfig.default"
versioninfo="versioninfo.txt"
buildinfo="$STAGINGDIR/buildconfig.info"

# Store some information for use in the build information of the image.
rm -f "${buildinfo}"
echo "BUILD_VER=`echo $curdir | sed -e 's/.*\/build\///'`" >> "${buildinfo}"
echo "BUILD_DEF=${prjdef}" >> "${buildinfo}"
echo "BUILD_START=${starttime}" >> "${buildinfo}"
echo "BUILD_HOST=`uname -n`" >> "${buildinfo}"
echo "BUILD_USER=$USER" >> "${buildinfo}"
echo "BUILD_CUSTOMER=$BUILD_CUSTOMER" >> "${buildinfo}"
echo "BUILD_PROJECT=$BUILD_PROJECT" >> "${buildinfo}"
echo "BUILD_HARDWARE=$BUILD_HARDWARE" >> "${buildinfo}"
echo "BUILD_CONFIGTYPE=$BUILD_CONFIGTYPE" >> "${buildinfo}"
echo "BUILD_SPIN_NUMBER=$spinnumber" >> "${buildinfo}"

if [ "${custom}" = "1" ]; then
  echo "-- Ready for customisation, press enter to start build."
  read
fi

opt_notrackws=""
if [ "${notrackws}" == "1" ]; then
  opt_notrackws="-n"
fi

opt_tracing=""
if [ "${enable_tracing}" == 1 ] ; then
  opt_tracing="-t"
fi

opt_createpackage=""
if [ "${createpackage}" == 1 ] ; then
  opt_createpackage="-z"
fi

opt_noforcesync=""
if [ "${noforcesync}" == "1" ]; then
  opt_noforcesync="-nf"
fi

if [ "${regendefconfig}" == "1" ]; then
  echo "-- Get source code"
    ./getsource.tcl $opt_notrackws $opt_noforcesync -s $servercfgfile -p $prjdep -r
  # Create fixed link to the output dir
  rootdir=`head -n1 rootpath.cfg | cut -d= -f2`
  rm -f $rootdir/output
  ln -s $outputdir $rootdir/output
elif [ "${nosync}" != "1" ]; then
  echo "-- Get source code"
    ./getsource.tcl $opt_notrackws $opt_noforcesync $opt_tracing $opt_createpackage -s $servercfgfile -p $prjdep
  # Create fixed link to the output dir
  rootdir=`head -n1 rootpath.cfg | cut -d= -f2`
  rm -f $rootdir/output
  ln -s $outputdir $rootdir/output
fi

if [ "${synconly}" = "1" ]; then
  endtime=`date +%Y-%m-%d_%H:%M:%S`
  echo "-- Synchronisation completed at ${endtime}"
  rm -rf "$outputdir"
  exit 0
fi

if [ -f $prjdir/test.config ]
then
  echo "test.config exists (copy $prjdir/test.config to $outputdir/test)"
  install -d $outputdir/test
  install -m 664 $prjdir/test.config $outputdir/test/
fi

if [ "${noconf}" != "1" ]; then
  echo "-- Configure build tree"

  # check for ccache presence
  ccache_where=`whereis ccache | wc -w`
  ccache_block=`grep "CONFIG_BUILD_NO_CCACHE=y" $prjdir/defconfig | wc -w`
  if [ "$ccache_where" -gt "1" ] && [ "${noccache}" != "1" ] && [ "$ccache_block" != "1" ]; then
    echo "-- Ccache is present: enabling"
    ccache_enable="-c"
  else
    ccache_enable=""
  fi

  # check for legacy releasedir support
  legacy_releasedir=`grep "CONFIG_BUILD_LEGACY_RELEASEDIR=y" $prjdir/defconfig | wc -w`
  if [ "$legacy_releasedir" == "1" ]; then
    echo "-- Legacy RELEASEDIR=ROOT_FS_DIR support enabled"
    legacy_releasedir="-y"
  else
    legacy_releasedir=""
  fi

  opt_rootfromvcs=""
  if [ "${rootfromvcs}" == "1" ]; then
    opt_rootfromvcs="-r"
  fi

  config_source_cmd="./configsource.tcl -s $servercfgfile -p $prjdep -h $hardcofile -o $outputdir -b $bldcfg $opt_rootfromvcs $ccache_enable $legacy_releasedir $opt_tracing $opt_createpackage"
  echo ${config_source_cmd}
  ${config_source_cmd}
  cat $versioninfo | sort > $STAGINGDIR/buildconfig.dep
  ./gentestconfig.tcl -s $servercfgfile -p $prjdep -h $hardcofile -o $outputdir -b $bldcfg $opt_rootfromvcs $ccache_enable $legacy_releasedir
fi

### Copy the project dir to the staging dir
install -d ${STAGINGDIR}/config
cp -aR $prjdir/* ${STAGINGDIR}/config
chmod -R +w ${STAGINGDIR}/config

### Install build configuration
echo "-- Install build configuration"
install -m 664 $prjdir/defconfig .config
install -m 664 .config ${STAGINGDIR}/components.config
make install-component-config

if [ "${doc}" = "1" ]; then
  echo "-- Build documentation"
  make doc
  endtime=`date +%Y-%m-%d_%H:%M:%S`
  echo "-- Building documentation completed at ${endtime}"
  echo "BUILD_END=${endtime}" >> "${buildinfo}"
  exit 0
fi

if [ "${tests}" = "1" ]; then
  echo "-- Build tests"
  make -i tests
  endtime=`date +%Y-%m-%d_%H:%M:%S`
  echo "-- Building tests completed at ${endtime}"
  echo "BUILD_END=${endtime}" >> "${buildinfo}"
  exit 0
fi

### Activate parallel build
JOBS=""
if [ "${fast}" = "1" ]; then
    CPUS=$(grep processor /proc/cpuinfo|wc -l)
    JOBS="-j$CPUS"
    echo "-- Fast mode activated (${CPUS} cpu(s))"
fi

if [ "${nobuild}" = "1" ]; then
  endtime=`date +%Y-%m-%d_%H:%M:%S`
  echo "-- Synchronisation and configuration completed at ${endtime}"
  echo "BUILD_END=${endtime}" >> "${buildinfo}"
  exit 0
fi


### Build all components
echo "-- Build all components"
make ${JOBS}

### Create image
echo "-- Create image"
make image


##
if [ "${createpackage}" == "1" ]; then
  echo "-- Copy the staging directory to the package directory"
  ./createPackageCopyStagingDir.sh
  echo "-- Tar the package directory"
  rm -rf output/release/delivery/Package.tar.gz
  tar -C output/package -zcf output/release/delivery/Package.tar.gz .
fi

### Build ready
endtime=`date +%Y-%m-%d_%H:%M:%S`
echo "-- Build completed at ${endtime}"
echo "BUILD_END=${endtime}" >> "${buildinfo}"


