#!/bin/bash
#
# This script is a wrapper for the install program.
# Its purpose is to transparantly install a file, with a given virtual path, into a specified file system layer. A layer of the filesystem corresponds
# to a set of files that have the same access and persistance properties.
# The path is considered virtual because it will not necessarily correspond to the path of the physical file in the run-time filesystem. Instead 
# the virtual path P might resolve to a symlink that points to the physical file that was placed under a "translated" root directory T at path T/P.
# Such a translated root directory is the physical form of an abstract "layer" of the filesystem. By putting all files of one layer under a common
# translated root directory we can move the entire layer onto a certain physical filesystem which can then be mounted by the system at run-time with the required
# layer-properties. There could also be multiple translated roots on one physical filesystem.
#
# The user need but specify the virtual path of the file and the layer in which it must be installed. The script takes care of setting up files, directories
# and symbolic links so that the build system can compile them into (an) image(s) that can be deployed onto a device where, at run-time, after some mounting,
# all files will be reachable through the specified virtual path and have properties correspoding to their specified layer.
#
# A layer can be chosen by passing an option to the script. This script imposes some constraints onto the relationship between the layer of a directory and
# the layer of its children. The layer of the directory must be higher or equal than that of its children according to the following ordering:
# --ro-persist-upgrade > --rw-persist-upgrade > --rw-persist-noupgrade > rw-ram
# It is also forbidden to install something directly into a mountpoint(we don't consider "/" to be a mountpoint) for obvious reasons.
#
# This has been done for several reasons:
#  - If all directories in a path could be in an arbitrary layer we would no longer be able to put all files of one layer under a single
#    transalated root directory T at path T/P, where P is their virtual path. For example, consider P=/a/b/c where a is on layer 1, b is on layer
#    2 and c on layer 1. Suppose t1 (translated root of layer 1) == "/T1" and t2 == /T2, then we can simply make a symlink "/a -> /T1/a" and mkdir /T1/a, 
#    /a/b would then be be a symlink in /T1/a/b -> /T2/a/b and we could again mkdir /T2/a/b, then /a/b/c would have to be a symlink /T2/a/b/c -> /T1/a/b/c
#    but then /T2/a/b/c == /T1/a/b/c so this is a problem. To solve this problem we could symlink /T2/a/b/c -> /T1/a/b/c_1 and put the file in /T1/a/b/c_1 but 
#    this would quickly lead to a bloated, complicated filesystem.
#  - If we assume that a higher layer is more secure than a lower layer, a file resolved by a link in a lower layer would only be as secure as the layer the 
#    link is in. This could be bad.
#
#  TODO: specify the layers in a seperate file and make a generic option that can be used to select the layer. We might also consider specifying the mountpoints
#        in that file so that the script can prohibit direct installations to a mountpoint.
# Errorprone test cases:
#  1) scripts/install/vfsl_install.sh Makefile.global Makefile.components /fs/rw/system --rw-persist-upgrade
#
# Get the script directory
MY_DIR=$(dirname $(readlink -f ${BASH_SOURCE[0]}))

# Set the type of operational disk (This wil become obsolete once we fully move to dual bank layout)
unset FIRMTYPE
if [ -z "$CONFIG_GENERATION_RESC" ] ; then
  FIRMTYPE=system
else
  FIRMTYPE=rescue
fi

usage() {
  echo "--
Wrapper options:
		--ro-persist-upgrade		# This installs files in the read only, persistent, upgradable part of the fs
		--rw-persist-upgrade		# This installs files in the read-writable, persistent, upgradable part of the fs
		--rw-persist-noupgrade		# This installs files in the read-writable, persistent part of the fs that will not be wiped during upgrade
		--rw-ram			# This installs files in the read-writable, volatile part of the filesystem
		--dry-run			# Only print the commands, don't run them
	-h	--help				
	-v	--version
" >&2
  install --help >&2
}

# Some definitions
TEMP_MOUNTPOINT=/var
RW_MOUNTPOINT=/mnt/jffs2/jffs2_3
RW_SYSTEM_ROOT=${RW_MOUNTPOINT}/system
RW_COMMON_ROOT=${RW_MOUNTPOINT}/common
   
REAL_SYSTEM_LOCATION=$(${MY_DIR}/resolve_virtual_path.sh "${RW_SYSTEM_ROOT}" )
REAL_COMMON_LOCATION=$(${MY_DIR}/resolve_virtual_path.sh "${RW_COMMON_ROOT}" )
REAL_TEMP_LOCATION=$(${MY_DIR}/resolve_virtual_path.sh "${TEMP_MOUNTPOINT}" )
REAL_ROOT_LOCATION=$(${MY_DIR}/resolve_virtual_path.sh "/" )

PROGNAME=${0##*/}
SHORTOPTS="hv"
LONGOPTS="writable,ro-persist-upgrade,rw-persist-upgrade,rw-persist-noupgrade,rw-temp,help,version,print,common,dry-run"

INSTALL_LONGOPTS="backup:,compare,directory,group:,mode:,owner:,preserve-timestamps,strip,strip-program:,suffix:,target-directory:,no-target-directory,verbose,preserve-context,context:"
INSTALL_SHORTOPTS="bcCdDg:m:o:psS:t:TvZ:"

ARGS=$(getopt -s bash --options ${SHORTOPTS}${INSTALL_SHORTOPTS}  \
  --longoptions "${LONGOPTS},${INSTALL_LONGOPTS}" --name $PROGNAME -- "$@" )

if [ "0" != "$?" ]; then
  echo "Argument parse error" >&2;
  exit $?;
fi

unset INSTALL_OPTIONS
unset MOUNTPOINT
unset PRINT
unset DRY_RUN
unset T_IS_SET
unset D_IS_SET

eval set -- "$ARGS"

# We handle our parameters and modify some of the install parameters. The rest will be passed to install
while true; do
   case $1 in
      -h|--help)
         usage
         exit 0
         ;;
      -v|--version)
         echo $PROGNAME: v0.1
	 ${INSTALL} --version
         exit 0
         ;;
      --ro-persist-upgrade) # root files
	MOUNTPOINT=/
         ;;
      --rw-temp) # temporary files
	if [ $MOUNTPOINT ]; then echo "A file can only be installed in one location"; exit 3; fi
	MOUNTPOINT=${TEMP_MOUNTPOINT}
	# Since no two files corresponding to different root_fs will exist at the same time it is not necessairy to put 
        # temp files in different directories depending on the active root filesystem
         ;;
      --rw-persist-upgrade) # rw-files for the operational system
	if [ $MOUNTPOINT ]; then echo "A file can only be installed in one location"; exit 3; fi
	MOUNTPOINT=${RW_SYSTEM_ROOT}
         ;;
      --rw-persist-noupgrade) # rw-files common to different root_fs banks of the operational system and the rescue system
	if [ $MOUNTPOINT ]; then echo "A file can only be installed in one location"; exit 3; fi
	MOUNTPOINT=${RW_COMMON_ROOT}
         ;;
      --print)
	 PRINT=1
         ;;
      --dry-run)
	 DRY_RUN=1
         ;;
      -d|--directory)
        D_IS_SET=1
	INSTALL_OPTIONS+=" --directory"
	;;
      -t|--target-directory)
        T_IS_SET=1
	T_ARGUMENT=$2
	if [  "${T_ARGUMENT:0:1}" != "/" ]; then
		echo "ERROR: argument ${T_ARGUMENT} is not an absolute path. " >&2
		exit 1;
	fi
        shift 1
	;;
      --)
        shift
	break
	;;
      --*)
	INSTALL_OPTIONS+=" $1"
         ;;
      -[^-])
	INSTALL_OPTIONS+=" $1"
         ;;
      *)
	INSTALL_OPTIONS+=" $1"
         ;;
   esac
   shift
done

ROOT_FS_DIR=$( echo ${ROOT_FS_DIR} | ${MY_DIR}/../file_tracing/reduce_path.awk )
ROOT_FS_DIR=${ROOT_FS_DIR%/} # We make sure the path does not end with "/"
# white box deduction
OUTPUT_DIR=$( echo ${ROOT_FS_DIR}/../.. | ${MY_DIR}/../file_tracing/reduce_path.awk )
OUTPUT_DIR=${OUTPUT_DIR%/} # We make sure the path does not end with "/"

# This is a hack so that the the install wrapper can be used as before with legacy release code
legacy_preprocess_args() {
    # This will cut the $ROOT_FS_DIR prefix, if any. In the past the real install was called where the target path was the full path
    # relative to ROOT_FS_DIR.
    ARG=$( echo $1 | ${MY_DIR}/../file_tracing/reduce_path.awk )
    if [[ ${ARG} == ${ROOT_FS_DIR}/* ]]; then # if we install in the root_fs
        ARG=${ARG##${ROOT_FS_DIR}}
    elif [[ ${ARG} == ${OUTPUT_DIR}/* ]]; then # if we install somewhere else in the outputdir (things are installed in staging during release ;-( ) 
        INSTALL_ARGS=${@:2:${#@}}
        verbose_install ${INSTALL_ARGS}
        exit 1
    else 
        ARG=$1
    fi
    echo $ARG
}

if ! [ $MOUNTPOINT ]; then
	MOUNTPOINT=/
fi

verbose_install() {
    echo install $@
    install $@
}
verbose_ln() {
    echo ln $@
    ln $@
}
verbose_mkdir() {
    echo mkdir $@
    mkdir $@
}

if [ $DRY_RUN ]; then
    INSTALL="echo install"
    LN="echo ln"
    MKDIR="echo mkdir"
else # of course this needs to be replaced by real commands
    INSTALL="verbose_install"
    LN="verbose_ln"
    MKDIR="verbose_mkdir"
fi

# Function to get the physical link location
get_physical_link_path() {
   
   BASE_PATH=${1}
   # This would be the physical location of the symlink if we were to put it in $(dirname $1)
   PHYS_LINK_PATH=$(${MY_DIR}/resolve_virtual_path.sh "${BASE_PATH}" )

#   echo "PHYS_LINK_PATH    =$PHYS_LINK_PATH" >&2
#   echo "REAL_ROOT_LOCATION=$REAL_ROOT_LOCATION" >&2
#   echo "REAL_RW_LOCATION  =$REAL_RW_LOCATION" >&2
#   echo "REAL_TEMP_LOCATION=$REAL_TEMP_LOCATION" >&2 

   # We verify that the user is not trying to make a link to a more constrained part of the filesystem.
   # This is the case if the physical location of the symlink is on a more constrained part of the filesystem
   # than the targeted part of the filesystem where the file is supposed to be placed.
   # We also check if we are not trying to make a symlink in the same location as the physical file. This can happen
   # if one of the parent directories is already on this mount_point. In this case no symlink should be made.
   # WARNING: Note that the comparison chain only works because the paths are implicitely ordered by decreasing length
   if [[ ${PHYS_LINK_PATH%/} == ${REAL_SYSTEM_LOCATION%/}/* ]]; then
#       echo "case 1" >&2
                if [ "${MOUNTPOINT}" == "${RW_SYSTEM_ROOT}" ]; then # NOK; cyclic symlink
                        PHYS_LINK_PATH="";
                elif [ "${MOUNTPOINT}" == "/" ]; then # NOK; 
                       echo "ERROR: $0: Links from the rw part of the filesysten to the ro-filesystem are forbidden: $(dirname "${BASE_PATH}") -> $(dirname "${PHYS_LINK_PATH}")" >&2;
                       exit 1;
                fi
   elif [[ ${PHYS_LINK_PATH%/} == ${REAL_COMMON_LOCATION%/}/* ]]; then
#       echo "case 1" >&2
                if [ "${MOUNTPOINT}" == "${RW_COMMON_ROOT}" ]; then # NOK; cyclic symlink
                        PHYS_LINK_PATH="";
                elif [ "${MOUNTPOINT}" == "${RW_SYSTEM_ROOT}" ]; then # NOK; 
                       echo "ERROR: $0: Links from the system part to the common part are forbidden: $(dirname "${BASE_PATH}") -> $(dirname "${PHYS_LINK_PATH}")" >&2;
                       exit 1;
                elif [ "${MOUNTPOINT}" == "/" ]; then # NOK; 
                       echo "ERROR: $0: Links from the rw part of the filesysten to the ro-filesystem are forbidden: $(dirname "${BASE_PATH}") -> $(dirname "${PHYS_LINK_PATH}")" >&2;
                       exit 1;
                fi
   elif [[ ${PHYS_LINK_PATH%/} == ${REAL_TEMP_LOCATION%/}/* ]]; then
#       echo "case 2" >&2
                if [ "${MOUNTPOINT}" == "${TEMP_MOUNTPOINT}" ]; then # NOK; cyclic symlink
                        PHYS_LINK_PATH="";
                elif [ "${MOUNTPOINT}" == "${RW_SYSTEM_ROOT}" ] || [ "${MOUNTPOINT}" == "${RW_COMMON_ROOT}" ] ; then # NOK;
                       echo "ERROR: $0: Links from the temp part of the filesysten to the rw-filesystem are forbidden. $(dirname "${BASE_PATH}") -> $(dirname "${PHYS_LINK_PATH}")" >&2; exit 1;
                elif [ "${MOUNTPOINT}" == "/" ]; then # NOK;
                       echo "ERROR: $0: Links from the temp part of the filesysten to the ro-filesystem are forbidden: $(dirname "${BASE_PATH}") -> $(dirname "${PHYS_LINK_PATH}")" >&2; exit 1;
                fi
   elif [[ ${PHYS_LINK_PATH%/} == ${REAL_ROOT_LOCATION%/}/* ]]; then
#       echo "case 3" >&2
       if [ "${MOUNTPOINT}" == "/" ]; then # NOK;
           PHYS_LINK_PATH="";
       fi
   else # At this point, since root is handled, the PHYS_LINK_PATH can only correspond to one of the MOUNTPOINTS
#       echo "ERROR: $0: You can not write a file to \"${BASE_PATH}\" since it is a mountpoint." >&2
#       exit 1
        PHYS_LINK_PATH="";
   fi

   echo $PHYS_LINK_PATH;
}

getMountedFilePath() {
    
    if [ "${BASE_PATH:0:1}" != "/" ]; then
        echo "ERROR: Argument ${BASE_PATH} is not an absolute path. " >&2
        exit 1;
    fi
    
    BASE_PATH=$( echo "$1" | ${MY_DIR}/../file_tracing/reduce_path.awk )
    if [[ ${BASE_PATH%/} == ${RW_MOUNTPOINT} ]] || [[ ${BASE_PATH%/} == ${TEMP_MOUNTPOINT} ]] || [[ ${BASE_PATH} == "/" ]]; then
        # We allow installing the mountpoints ( options like --ro-persist-upgrade are ignored in this case )
        MOUNTED_FILE_PATH="${BASE_PATH}"
    elif [[ ${BASE_PATH%/}/ == ${RW_MOUNTPOINT}/* ]] || [[ ${BASE_PATH%/}/ == ${TEMP_MOUNTPOINT}/* ]]; then 
        echo "ERROR: $0: Installing a file directly inside a mount point is forbidden." >&2;
        exit 1;
    else
        # Any other path falls under a mountpoint
        MOUNTED_FILE_PATH="${MOUNTPOINT%/}${BASE_PATH}"
    fi

    echo $MOUNTED_FILE_PATH
}

get_install_path() {
    BASE_PATH=$1

    if [ "${BASE_PATH:0:1}" != "/" ]; then
        echo "ERROR: Argument ${BASE_PATH} is not an absolute path." >&2
        exit 1;
    fi

    # This is the path of the physical file on the fully mounted filesystem
    MOUNTED_FILE_PATH=$(getMountedFilePath ${BASE_PATH}); if ! [ $? -eq 0 ]; then exit 1; fi;

    # This is the path of the physical file in the build filesystem
    # The rest of the build system will make sure it will get mounted in the right place in the operational system
    # Basically the script deduces the real path based on mount-point information
#    echo "${MY_DIR}/resolve_virtual_path.sh \"${MOUNTED_FILE_PATH}\"" >&2
    BUILD_FS_PATH=$(${MY_DIR}/resolve_virtual_path.sh "${MOUNTED_FILE_PATH}" ); if ! [ $? -eq 0 ]; then exit 1; fi;
   
    echo ${BUILD_FS_PATH}
}

get_symlink_to_target() {
    BASE_PATH=$1

    # This is the path of the physical symlink on the ROOT_FS
    # Basically this call deduces the real path based on mount-point information and symlinks already present
    BUILD_FS_LINK_PATH=$( get_physical_link_path "${BASE_PATH}" ); if ! [ $? -eq 0 ]; then exit 1; fi;
    if ! [ $? -eq 0 ]; then exit 1; fi;
       
    # If the path is already a link 
    if [ ! -z "${BUILD_FS_LINK_PATH}" ]; then
        if [ -e ${BUILD_FS_LINK_PATH} ]; then
          if [ -L ${BUILD_FS_LINK_PATH} ]; then
              echo "ERROR: $0: Internal error(FIXME): symlink ${BUILD_FS_LINK_PATH} should have been fully resolved. " >&2
          else
              echo "ERROR: $0: ${BASE_PATH} already exists. " >&2
          fi
          exit 1;
        elif ! [ $D_IS_SET ] && ! [ -d $(dirname ${BUILD_FS_LINK_PATH}) ]; then
            # if D is SET than we can create the directory, just as install would if we were installing a file instead of a symlink
            echo "ERROR: $0: $(dirname ${BASE_PATH}) -> $(dirname ${BUILD_FS_LINK_PATH}) is not a directory." >&2
            exit 1;
        fi
    fi
    echo $BUILD_FS_LINK_PATH
}

make_symlink_to_target() {
    BASE_PATH=$1
#    echo "make_symlink_to_target: $BASE_PATH" >&2

    if [ "${BASE_PATH:0:1}" != "/" ]; then
        echo "ERROR: Argument ${BASE_PATH} is not an absolute path. " >&2
        exit 1;
    fi
    # This is the path of the physical file on the fully mounted filesystem
    MOUNTED_FILE_PATH=$(getMountedFilePath ${BASE_PATH}); if ! [ $? -eq 0 ]; then exit 1; fi;

    # This is the path of the physical symlink on the ROOT_FS
    # Basically this call deduces the real path based on mount-point information and symlinks already present
    BUILD_FS_LINK_PATH=$( get_symlink_to_target ${BASE_PATH} ); if ! [ $? -eq 0 ]; then exit 1; fi;


    if [ $? -eq 0 ]; then
        if ! [ -z ${BUILD_FS_LINK_PATH} ]; then
	    BUILD_FS_LINK_DIR=$(dirname "${BUILD_FS_LINK_PATH}" )
            if ! [ -d ${BUILD_FS_LINK_DIR} ]; then 
                # get_symlink_to_target verifies that in this case D_IS_SET must be set
                ${MKDIR} -p ${BUILD_FS_LINK_DIR};
            fi; 
            ${LN} -s ${MOUNTED_FILE_PATH} ${BUILD_FS_LINK_PATH}
	fi
    else
	exit $?
    fi;
}

if [ $PRINT ]; then
       echo $( get_install_path ${MOUNTPOINT} );
       exit 0;
fi

if ! [ -d $REAL_SYSTEM_LOCATION ]; then
    ${MKDIR} -p $REAL_SYSTEM_LOCATION
fi
if ! [ -d $REAL_COMMON_LOCATION ]; then
    ${MKDIR} -p $REAL_COMMON_LOCATION
fi
if ! [ -d $REAL_TEMP_LOCATION ]; then
    ${MKDIR} -p $REAL_TEMP_LOCATION
fi
if ! [ -d $REAL_ROOT_LOCATION ]; then
    ${MKDIR} -p $REAL_ROOT_LOCATION
fi

if [ $T_IS_SET ]; then # the argument of option -t is the destination
    # Install only allows the source arguments to be files

    T_ARGUMENT=$(legacy_preprocess_args ${T_ARGUMENT} $INSTALL_OPTIONS --target-directory=${T_ARGUMENT} $@ ); if [ $? -eq 1 ]; then exit 0; fi;
    # We first make the symlink because this function verifies that the user
    # is not trying to put a file in a more constrained part of the fs while it's parent directory
    # is on a less constrained part of the fs. Otherwise we would get complicated links situations.
    make_symlink_to_target ${T_ARGUMENT}; if ! [ $? -eq 0 ]; then exit 1; fi;

    FULL_BUILD_FS_PATH=$( get_install_path ${T_ARGUMENT} ); if ! [ $? -eq 0 ]; then exit $?; fi;

    INSTALL_OPTIONS+=" --target-directory=${FULL_BUILD_FS_PATH}"

            # the target is a directory so for every source that will go into the directory we might need to make a link
            for ((i=1;i<=$#;++i)); do
                # This checks if every link can be made
		FILENAME=$( basename ${!i} )
                BLA=$( get_symlink_to_target "${T_ARGUMENT}/${FILENAME}" ); if ! [ $? -eq 0 ]; then exit 1; fi;
                ARGUMENTS+=" ${!i}"
            done

            # We need to create any missing directories in the path of the physical file. They might not exist yet
            # since it is possible that a file of this type has not been created yet in that path.
            FULL_BUILD_FS_DIR=$(dirname ${FULL_BUILD_FS_PATH})
            if ! [ -d ${FULL_BUILD_FS_DIR} ]; then
                ${MKDIR} -p ${FULL_BUILD_FS_DIR}
            fi
            ${INSTALL} ${INSTALL_OPTIONS} ${ARGUMENTS}; if ! [ $? -eq 0 ]; then exit 1; fi;

            # If the install doesn't fail we make the actual links
            for ((i=1;i<=$#;++i)); do
		FILENAME=$( basename ${!i} )
                make_symlink_to_target "${T_ARGUMENT}/${FILENAME}"; if ! [ $? -eq 0 ]; then exit 1; fi;
            done

    
elif [ $D_IS_SET ]; then # All arguments are directories that need to be created

    # We first check if all links can be created
    for DIR in $@; do
        DIR=$(legacy_preprocess_args ${DIR} $INSTALL_OPTIONS $@); if [ $? -eq 1 ]; then exit 0; fi;
        BLA=$( get_symlink_to_target "${DIR}" ); if ! [ $? -eq 0 ]; then exit $?; fi;
        FULL_BUILD_FS_PATH=$( get_install_path ${DIR} ); if ! [ $? -eq 0 ]; then exit $?; fi;
        ARGUMENTS+=" ${FULL_BUILD_FS_PATH}"
    done
    
    ${INSTALL} ${INSTALL_OPTIONS} ${ARGUMENTS}; if ! [ $? -eq 0 ]; then exit 1; fi;
    
    for DIR in $@; do
        DIR=$(legacy_preprocess_args ${DIR} $INSTALL_OPTIONS $@); if [ $? -eq 1 ]; then exit 0; fi;
    	make_symlink_to_target ${DIR}; if ! [ $? -eq 0 ]; then exit 1; fi;
    done
    
else # the last argument is the destination
	
    if [  $# -lt 2 ]; then
	echo "ERROR: need at least two arguments when -t nor -d is set" >&2
	usage
	exit 1;
    else
        TARGET_PATH=$(legacy_preprocess_args ${!#} $INSTALL_OPTIONS $@); if [ $? -eq 1 ]; then exit 0; fi;
        FULL_BUILD_FS_PATH=$( get_install_path ${TARGET_PATH} ); if ! [ $? -eq 0 ]; then exit $?; fi;
        if [ -d ${FULL_BUILD_FS_PATH} ] || [[ $TARGET_PATH == */ ]]; then
	   TARGET_PATH=$( echo "${TARGET_PATH}" | sed 's#/*$##' )
#           echo "$0: ${FULL_BUILD_FS_PATH} is directory" >&2
            # the target is a directory so for every source that will go into the directory we might need to make a link
            for ((i=1;i<$#;++i)); do
                # This checks if every link can be made
		FILENAME=$( basename ${!i} )
                BLA=$( get_symlink_to_target "${TARGET_PATH}/${FILENAME}" ); if ! [ $? -eq 0 ]; then exit 1; fi;
                ARGUMENTS+=" ${!i}"
            done
            ARGUMENTS+=" ${FULL_BUILD_FS_PATH}"

            # We need to create any missing directories in the path of the physical file
            if ! [ -d ${FULL_BUILD_FS_PATH} ]; then
                ${MKDIR} -p ${FULL_BUILD_FS_PATH}
            fi
            ${INSTALL} ${INSTALL_OPTIONS} ${ARGUMENTS}; if ! [ $? -eq 0 ]; then exit 1; fi;
            # If the install doesn't fail we make the actual links
            for ((i=1;i<$#;++i)); do
		FILENAME=$( basename ${!i} )
    		make_symlink_to_target "${TARGET_PATH}/${FILENAME}"; if ! [ $? -eq 0 ]; then exit 1; fi;
            done
	else
 #          echo "$0: ${FULL_BUILD_FS_PATH} is not a directory" >&2
              # We can assume that there is only one source argument since installing multiple files to one file 
              # will cause install to return an error
              BLA=$( get_symlink_to_target "${TARGET_PATH}" ); if ! [ $? -eq 0 ]; then exit 1; fi;
              for ((i=1;i<$#;++i)); do
                  ARGUMENTS+=" ${!i}"
              done
              ARGUMENTS+=" ${FULL_BUILD_FS_PATH}"

              # We need to create any missing directories in the path of the physical file
              FULL_BUILD_FS_DIR=$(dirname ${FULL_BUILD_FS_PATH})
              if ! [ -d ${FULL_BUILD_FS_DIR} ]; then
                  ${MKDIR} -p ${FULL_BUILD_FS_DIR}
              fi
              ${INSTALL} ${INSTALL_OPTIONS} ${ARGUMENTS}; if ! [ $? -eq 0 ]; then exit 1; fi;
              # If the install doesn't fail we make the actual links
              make_symlink_to_target "${TARGET_PATH}"; if ! [ $? -eq 0 ]; then exit 1; fi;
        fi
    fi
fi;
