/*
 * ============================================================================
 *
 * Copyright : (c) 2005 SAGEM SA - Site Velizy - URD2
 *
 * The information and source code contained herein is the exclusive
 * property of SAGEM and may not be disclosed, examined, or
 * reproduced in whole or in part without explicit written authorization
 * of the company.
 *
 * ============================================================================
 */

/**
 * \file bsp_dev.c Debug device process
 *
 * This file contain all BSP device process.
 */

/*
 * Includes
 * ~~~~~~~~
 */

#include "bsp_pri.h"

/*
 * Functions
 * ~~~~~~~~~
 */

/** Open device.
 *
 *   \par Description
 *	<b>This function allow to perform open command.</b>\n
 *
 *	@param _pInode:	Inode ioctl.
 *	@param _pFile:		file ioctl.
 *
 *	@return on success 0, on error -1.
 */
int bsp_devOpen(	struct inode	*_pInode,
				struct file	*_pFile)
{
	unsigned int _minor = MINOR(_pFile->f_dentry->d_inode->i_rdev);

	/* If bsp */
	HALBSP_DEBUG(	"Begin");

	/* Semaphore */
	if (down_interruptible(&bsp_base.Dev[_minor].sem)) {
		/* Signal */
		return(-ERESTARTSYS);
	}
	
	/* If last User Land application open file descriptor */
	if (bsp_base.Dev[_minor].nbOpen == 0) {
	}

	/* Increase Nb Open */
	(bsp_base.Dev[_minor].nbOpen)++;

	/* If bsp */
	HALBSP_DEBUG(	"nb Open(%d)", bsp_base.Dev[_minor].nbOpen);

	/* Semaphore */
	up(&(bsp_base.Dev[_minor].sem));

	/* If bsp */
	HALBSP_DEBUG(	"End");

	return(0);
}

/** Release device.
 *
 *   \par Description
 *	<b>This function allow to perform release command.</b>\n
 *
 *	@param _pInode:	Inode ioctl.
 *	@param _pFile:		file ioctl.
 *
 *	@return on success 0, on error -1.
 */
int bsp_devRelease(	struct inode	*_pInode,
				struct file	*_pFile)
{
	unsigned int _minor = MINOR(_pFile->f_dentry->d_inode->i_rdev);

	/* If bsp */
	HALBSP_DEBUG(	"Begin");

	/* Semaphore */
	if (down_interruptible(&bsp_base.Dev[_minor].sem)) {
		/* Signal */
		return(-ERESTARTSYS);
	}

	/* If bsp */
	HALBSP_DEBUG(	"nb Open(%d)", bsp_base.Dev[_minor].nbOpen);

	/* Decrease Nb open */
	(bsp_base.Dev[_minor].nbOpen)--;

	/* If last User Land application open file descriptor */
	if (bsp_base.Dev[_minor].nbOpen == 0) {
	}

	/* Semaphore */
	up(&(bsp_base.Dev[_minor].sem));

	/* If bsp */
	HALBSP_DEBUG(	"End");

	return(0);
}

/** Ioctl device.
 *
 *   \par Description
 *	<b>This function allow to perform ioctl command.</b>\n
 *
 *	@param _pFile:		file ioctl.
 *	@param _pBuffer: 	Reference to Buffer.
 *	@param _count:		Bytes to read.
 *	@param _pPos:		Position to read.
 *
 *	@return on success 0, on error -1.
 */
int bsp_devIoctl(	struct inode	*_pInode,
				struct file	*_pFile,
				unsigned int	_cmd,
				unsigned long	_arg)
{
	unsigned int _minor = MINOR(_pFile->f_dentry->d_inode->i_rdev);
	int			_ret	= 0;

	/* If bsp */
	HALBSP_DEBUG(	"Begin");

	/* Semaphore */
	if (down_interruptible(&bsp_base.Dev[_minor].sem)) {
		/* Signal */
		return(-ERESTARTSYS);
	}

	/* Process */
	{
		enum bsp_get_permParam_enu	_get_permParam	= _IOC_NR(_cmd);
		int						_size		= _IOC_SIZE(_cmd);
		void						*_pValue		= (void *)0;
		
		/* Check */
		if (_get_permParam >= BSP_Get_permParam_max) {
			HALBSP_ERROR(	"FAILURE -> cmd(%u) unknown",
					_cmd);
	
			/* Semaphore */
			up(&bsp_base.Dev[_minor].sem);              					
	
			/* Ko */
			return(-1);
		}

		/* Perform get permParam */
		_pValue = bsp_get_permParam(_get_permParam);

		/* Copy to User */
		if (copy_to_user(	(void *)_arg,
						_pValue,
						_size)) {
			HALBSP_ERROR(	"FAILURE -> Copy to user");
	
			/* Semaphore */
			up(&bsp_base.Dev[_minor].sem);              					
	
			/* Ko */
			return(-1);
		}
	}

	/* Semaphore */
	up(&bsp_base.Dev[_minor].sem);              					

	/* If bsp */
	HALBSP_DEBUG("End");

	return(_ret);
}

// vim:ts=5:
