/*
 * (c) Inventel s.a.
 * (c) Thomson multimedia
 * (c) 2008-2009 SoftAtHome
 *
 * This product is distributed under the terms of
 * the GNU General Public License, in which case the provisions of the GPL are
 * required INSTEAD OF the above restrictions.  (This clause is
 * necessary due to a potential bad interaction between the GPL and
 * the restrictions contained in a BSD-style copyright.)
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, ALL OF
 * WHICH ARE HEREBY DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF NOT ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include <linux/module.h>
#include <linux/poll.h>
#include <linux/wait.h>
#include <linux/time.h>
#include <linux/interrupt.h>
#include <linux/spinlock.h>
#include <linux/cdev.h>
//#include <linux/devfs_fs_kernel.h>
static inline int devfs_mk_cdev(dev_t dev, umode_t mode, const char *fmt, ...) { return 0; }
static inline void devfs_remove(const char *fmt, ...) { }
#include <linux/reboot.h> /* machine_restart */

#include <asm/uaccess.h>

#include "button_local.h"
#include "lb_button.h"

#ifdef HARDCO_BUTTON_DEBUG
#define DEBUG(format, args...)	printk(format, ##args)
#else
#define DEBUG(format, args...)
#endif

struct lb_trigger;
struct lb_button;

/* keep it a power of 2 */
#define LBBUTTON_MAX_EVENTS	128
#define LBBUTTON_NDEVS		kButtonEnd
#define LBBUTTON_NEVENTS	evButtonEnd
#define LBBUTTON_FREQ		10

struct lb_button_file {
	int rdx, wdx;
	struct lb_button_event events[LBBUTTON_MAX_EVENTS];

	struct fasync_struct *asyncq;   /* async users */
	wait_queue_head_t readq;        /* read waiters */

	struct lb_trigger *trigger;	/* one-to-one relation (so far) */
};

struct lb_trigger {
	struct list_head t_list;	/* link on button's trigger list */

	int (*do_action)(struct lb_trigger *, struct lb_button *);

	unsigned long param1, param2;	/* trigger-specific parameters */

	struct lb_button_file *file;
};

struct lb_button {
	int active;			/* 1:0, monitored or not */

	BOARD_BUTTON_NAME name;		/* button id */

	BOARD_BUTTON_STATE state, prev; /* current and previous states */

	spinlock_t lock;

	/* per-event-type lists of triggers watching the button */
	struct list_head triggers[LBBUTTON_NEVENTS];
};

static int lbbutton_asyncsig = SIGIO;
static struct timer_list lb_timer;

/* Initialize the name in required order to map minor device to GPIO pin */
static struct lb_button lb_buttons[LBBUTTON_NDEVS] =
{
	[0] = { .name = kButtonPairing },
	[1] = { .name = kButtonReboot },
	[2] = { .name = kButtonReset },
	[3] = { .name = kButtonWifiOn },
	[4] = { .name = kButtonWakeupDisplay },

};

static int __copy_events_localy(struct lb_button_file *bfile, void *buf, int nevents)
{
	int count;

	DEBUG("__copy_events_localy: nevents(%d), bfile(%x), wdx(%d), rdx(%d)\n",
			nevents, bfile, bfile->wdx, bfile->rdx);

	/* There is only one reader at any time. */
	count = bfile->wdx - bfile->rdx;
	
	if (!count) {
		DEBUG("__copy_events_localy: no data\n");
		return 0;
	} else 	if (count < 0)
		count = LBBUTTON_MAX_EVENTS - bfile->rdx;

	count = min_t(int, nevents, count);

	if (copy_to_user(buf, &bfile->events[bfile->rdx], count * sizeof(struct lb_button_event)))
		return -EFAULT;

	bfile->rdx = (bfile->rdx + count) & (LBBUTTON_MAX_EVENTS - 1);
	
	DEBUG("__copy_events_localy: ret(%d)\n", count);
	return count;
}

static int post_event(struct lb_trigger *trigger, BOARD_BUTTON_EVENT ev)
{
	struct lb_button_file *bfile = trigger->file;

	DEBUG("post_event: event(%d)\n", ev);

	/*
	 * Lockless circular buffer management.
	 *
  	 * There is only one writter at any time, so we normally can't run
 	 * in mess with wdx.
 	 *
	 * Yes, readers should consume events fast enough until we overwrite them.
 	 */
	
	bfile->events[bfile->wdx].event = ev;
	do_gettimeofday(&bfile->events[bfile->wdx].stamp);

	bfile->wdx = (bfile->wdx + 1) & (LBBUTTON_MAX_EVENTS - 1);

	DEBUG("post_event: bfile(%x), wdx(%d), rdx(%d)\n", bfile, bfile->wdx, bfile->rdx);

	wake_up_interruptible(&bfile->readq);
	kill_fasync(&bfile->asyncq, lbbutton_asyncsig, POLL_IN);
	
	return 0;
}

static int btn_action_pushed(struct lb_trigger *trigger, struct lb_button *button)
{
	if (button->state != kButtonPushed)
		return 0;

	post_event(trigger, evButtonPushed);
	return 0;
}

static int btn_action_released(struct lb_trigger *trigger, struct lb_button *button)
{
	if (button->state != kButtonReleased)
		return 0;

	post_event(trigger, evButtonReleased);
	return 0;
}

static int btn_action_kept_pushed(struct lb_trigger *trigger, struct lb_button *button)
{
	if (trigger->param2) {
		if (!(--trigger->param2))
			post_event(trigger, evButtonKeptPushed);
		else if (button->state == kButtonReleased)
			trigger->param2 = 0;
	} else if (button->state == kButtonPushed 
			&& button->prev == kButtonReleased)
		trigger->param2 = trigger->param1;

	return 0;
}

static int lbbutton_add_trigger(struct lb_button *btn,
				struct lb_button_trigger *tinfo,
				struct lb_button_file *bfile)
{
	struct lb_trigger *trigger;

	DEBUG("add_trigger: type(%d), data(%lu)\n", tinfo->event, tinfo->data);
	
	/* keep one-to-one relation (can be changed later) */
	if (bfile->trigger)
		return -EBUSY;

	trigger = kmalloc(sizeof(struct lb_trigger), GFP_KERNEL);
	if (!trigger)
		return -ENOMEM;
	memset(trigger, 0, sizeof(struct lb_trigger));
	INIT_LIST_HEAD(&trigger->t_list);

	trigger->param1 = 0;
	trigger->param2 = 0;

	switch (tinfo->event) {
		case evButtonPushed:
			trigger->do_action = btn_action_pushed;
			break;
		case evButtonReleased:
			trigger->do_action = btn_action_released;
			break;
		case evButtonKeptPushed:
			trigger->do_action = btn_action_kept_pushed;
			trigger->param1 = tinfo->data * LBBUTTON_FREQ;
			break;
		default:
			return -EINVAL;
	}

	bfile->trigger = trigger;
	trigger->file = bfile;

	spin_lock_bh(&btn->lock);
	list_add(&trigger->t_list, &btn->triggers[tinfo->event]);
	spin_unlock_bh(&btn->lock);

	DEBUG("add_trigger: done\n");
	return 0;
}

static int lbbutton_del_trigger(struct lb_button *btn,
				struct lb_button_trigger *tinfo,
				struct lb_button_file *bfile)
{
	DEBUG("del_trigger\n");
	
	if (!bfile->trigger)
		return -EINVAL;

	spin_lock_bh(&btn->lock);
	list_del(&bfile->trigger->t_list);
	spin_unlock_bh(&btn->lock);

	kfree(bfile->trigger);
	bfile->trigger = NULL;

	return 0;
}

static inline void __handle_button_triggers(struct list_head *list, struct lb_button *button)
{
	struct list_head *head;

	spin_lock(&(button->lock));
	
	list_for_each(head, list) {
		struct lb_trigger *trigger = list_entry(head, struct lb_trigger, t_list);
		trigger->do_action(trigger, button);
	}
	spin_unlock(&(button->lock));
}

static void handle_button_triggers(struct lb_button *button)
{
	button->prev  = button->state;	
	button->state = kbutton_hw_get_state(button->name);

	if (button->state != button->prev)
	{
		if (button->state == kButtonPushed)
		{
			__handle_button_triggers(&(button->triggers[evButtonPushed]), button);
#if defined(CONFIG_BUTTON_SOFT_REBOOT)
#if 0 /* Sagemcom-Akka(S.VIE), 2011-08-23: Could not find external symbol "machine_restart" */
			if(kButtonReboot == button->name)
				machine_restart(NULL);
#endif
#endif
		}
		else
			__handle_button_triggers(&(button->triggers[evButtonReleased]), button);	
	}
	__handle_button_triggers(&(button->triggers[evButtonKeptPushed]), button);
}

/* Generic callback. */
static void generic_button_callback(unsigned long data)
{
	int i = 0;

	/* Run all registered triggers. */
	for ( ; i < LBBUTTON_NDEVS; i++) {
		if (lb_buttons[i].active)
			handle_button_triggers(&lb_buttons[i]);
	}						

	lb_timer.expires = jiffies + HZ/LBBUTTON_FREQ;
	add_timer(&lb_timer);
}

/************************************************
 *		File operations.		*
 ************************************************/

static int lbbutton_open(struct inode *inode, struct file *file)
{
	int minor = MINOR(inode->i_rdev);
	struct lb_button_file *bfile;

	if (minor >= LBBUTTON_NDEVS) {
		/* hey, stop playing with mknod :) */
		DEBUG("lbbutton_open: minor(%d) >= max(%d)\n", minor, LBBUTTON_NDEVS);
		return -ENXIO;
	}

	bfile = kmalloc(sizeof(struct lb_button_file), GFP_KERNEL);
	if (!bfile) return -ENOMEM;
	memset(bfile, 0, sizeof(struct lb_button_file));

	DEBUG("lbbutton_open: bfile(%x)\n", bfile);
	
	init_waitqueue_head(&bfile->readq);
	
	file->private_data = bfile;

	return 0;
}

/* 
 * So far, we imply that a user has deleted all the triggers via ioctl()
 * prio to this step. This way, all waiters are informed.
 */
static int lbbutton_release(struct inode *inode, struct file *file)
{
	int minor = MINOR(inode->i_rdev);

	down(&inode->i_mutex);
	
	lbbutton_del_trigger(&lb_buttons[minor], NULL, file->private_data);
	kfree(file->private_data);

	/* The very first thing waiters have to do is to check file->private_data.
   	 * In case of NULL, the object has been "closed". */
	file->private_data = NULL;

	/* Bah, that's not safe... Be sure you don't have waiters at this point (so far) */
#if 0
	/* Waiters first will be blocked on i_mutex.
	 * So the order wrt the previous op. does not matter. */
	wake_up_interruptible_sync(&bfile->readq);
	kill_fasync(&bfile->asyncq, lbbutton_asyncsig, POLL_IN);
#endif	
	up(&inode->i_mutex);

	DEBUG("lbbutton_release: done\n");
	return 0;
}

/* Drop the inode semaphore and wait for a button event, atomically */
static int lbbutton_event_wait(struct inode * inode, struct lb_button_file *bfile)
{
	int sigpending;
	DEFINE_WAIT(wait);

	prepare_to_wait(&bfile->readq, &wait, TASK_INTERRUPTIBLE);
	up(&inode->i_mutex);
	schedule();
	finish_wait(&bfile->readq, &wait);
	sigpending = signal_pending(current);
	down(&inode->i_mutex);

	return sigpending;
}

static ssize_t lbbutton_read(struct file *file,
			     char *buf, size_t count, loff_t *ppos)
{
	struct inode *inode = file->f_dentry->d_inode;
	struct lb_button_file *bfile;
	int sigpending = 0;
	int ret;

	DEBUG("lbbutton_read: enter, count(%d)\n", count);
	
	/* is there a room for at least an event? */
	if (count < sizeof(struct lb_button_event))
		return -EINVAL;
	
	down(&inode->i_mutex);
	bfile = file->private_data;

	for (;;) {
		ret = __copy_events_localy(bfile, buf, (count / sizeof(struct lb_button_event)));

		if (ret)
			break;
		
		if (file->f_flags & O_NONBLOCK) {
			ret = -EAGAIN;
			break;
		}

		if (sigpending) {
			ret = -ERESTARTSYS;
			break;
		}	

		sigpending = lbbutton_event_wait(inode, bfile);
	}
	up(&inode->i_mutex);

	DEBUG("lbbutton_read: %d events\n", ret);
	return (ret > 0 ? ret * sizeof(struct lb_button_event) : ret);
}

static int lbbutton_ioctl(struct inode *inode,
			  struct file *file, unsigned int cmd, unsigned long arg)
{
	struct lb_button_trigger trigger;
	struct lb_button_file *bfile = NULL;
	int ret, minor;

	minor = MINOR(inode->i_rdev);

	if (minor >= LBBUTTON_NDEVS)
		return -ENXIO;

	DEBUG("lbbutton_ioctl: cmd(%d)\n", cmd);
	
	switch (cmd) {
		case LBBUTTON_ADD_TRIGGER:
			down(&inode->i_mutex);
			bfile = (struct lb_button_file *)file->private_data;

			if (copy_from_user(&trigger, (void*)arg, sizeof(trigger)))
				ret = -EFAULT;
			else 
				ret = lbbutton_add_trigger(&lb_buttons[minor], &trigger, bfile);
			
			up(&inode->i_mutex);
			break;

		case LBBUTTON_REMOVE_TRIGGER:
			down(&inode->i_mutex);
			bfile = (struct lb_button_file *)file->private_data;
			ret = lbbutton_del_trigger(&lb_buttons[minor], NULL, bfile);
			up(&inode->i_mutex);
			break;

		case LBBUTTON_GET_STATE:
			ret = kbutton_hw_get_state(lb_buttons[minor].name);
			break;
		default:
			ret = -EINVAL;
	}

	return ret;
}

static unsigned lbbutton_poll(struct file *file, poll_table *pt)
{
	struct lb_button_file *bfile;
	unsigned r_mask = 0;

	DEBUG("lbbutton_poll: enter\n");
	
	bfile = (struct lb_button_file *)file->private_data;
	poll_wait(file, &bfile->readq, pt);
	
	if (bfile->rdx != bfile->wdx)
		r_mask |= (POLLIN | POLLRDNORM);
	
	DEBUG("lbbutton_poll: r_mask(%x)\n", r_mask);
	return r_mask;
}

static int lbbutton_fasync(int fd, struct file *file, int on)
{
	struct inode *inode = file->f_dentry->d_inode;
	struct lb_button_file *bfile;
	int ret;

	DEBUG("lbbutton_fasync: on(%d)\n", on);
	
	down(&inode->i_mutex);
	bfile = (struct lb_button_file *)file->private_data;
	ret = fasync_helper(fd, file, on, &bfile->asyncq);
	up(&inode->i_mutex);

	if (ret < 0)
		return ret;

	return 0;
}

static struct file_operations lbbutton_fops = {
	.owner    = THIS_MODULE,
	.read     = lbbutton_read,
	.poll     = lbbutton_poll,
	.ioctl    = lbbutton_ioctl,
	.open     = lbbutton_open,
	.release  = lbbutton_release,
	.fasync   = lbbutton_fasync
};

static struct cdev lb_cdev;
static dev_t lb_dev;

static int __init lb_buttons_init(void)
{
	int err, i, j;

	err = alloc_chrdev_region(&lb_dev, 0, LBBUTTON_NDEVS, "lb_button");
	if (err) {
		printk(KERN_ERR "[ ERROR ] : alloc_chrdev_region() failed for LB Button API module\n");
		return -1;
	}

	cdev_init(&lb_cdev, &lbbutton_fops);
	lb_cdev.owner = THIS_MODULE;
	err = cdev_add(&lb_cdev, lb_dev, LBBUTTON_NDEVS);
	
	if (err < 0) {
		cdev_del(&lb_cdev);
		unregister_chrdev_region(lb_dev, LBBUTTON_NDEVS);
		printk(KERN_ERR "[ ERROR ] : cdev_add() failed for LB Button API module\n");
		return -1;
	}
	kbutton_hw_init();
	for (i = 0; i < LBBUTTON_NDEVS; i++) {
		lb_buttons[i].active = 1;
		devfs_mk_cdev(MKDEV(MAJOR(lb_dev), MINOR(lb_dev) + i),
			S_IFCHR|S_IRUSR|S_IWUSR|S_IRGRP, "button%d", i);
		lb_buttons[i].state = kbutton_hw_get_state(lb_buttons[i].name);
		lb_buttons[i].prev  = lb_buttons[i].state;

		spin_lock_init(&lb_buttons[i].lock);

		for (j = 0; j < LBBUTTON_NEVENTS; j++)
			INIT_LIST_HEAD(&lb_buttons[i].triggers[j]);
	}
	
	init_timer(&lb_timer);

	lb_timer.function = &generic_button_callback;
	lb_timer.data = 0;
	lb_timer.expires = jiffies + HZ/LBBUTTON_FREQ;

	add_timer(&lb_timer);

	printk(KERN_INFO "SoftAtHome Button driver has been initialized\n");
	return 0;
}

static void __exit lb_buttons_exit(void)
{
	int i;

	del_timer_sync(&lb_timer);

	for (i = 0; i < LBBUTTON_NDEVS; i++) {
		devfs_remove("button%d", i);
	}
	unregister_chrdev_region(lb_dev, LBBUTTON_NDEVS);
	cdev_del(&lb_cdev);
}

module_init(lb_buttons_init);
module_exit(lb_buttons_exit);
MODULE_LICENSE("GPL");
