#include <linux/init.h>
#include <linux/fs.h>
#include <asm/uaccess.h>

#include "khco_led_api.h"
#include "led_local.h"

extern int kSysLedCtrl(int, int);

// key structure of led and GPIO mapping
typedef struct _kled_map_pair_t
{
	BOARD_LED_NAME	name;		/* virtual led name */	
	unsigned int	mask;		/* physical GPIO pin mask */
	BOARD_LED_STATE initstate;	/* initial state */
	unsigned int	activelow;	/* reset bit to turn on LED */
	unsigned int	maskfail;	/* physical GPIO pin mask for state failure */
	unsigned int	activelowfail;	/* reset bit to turn on LED */
} kled_pair_map_t, *pkled_pair_map_t;

// led infomation structure
typedef struct _kled_info_t
{
	unsigned long mask;		// mask for led: ie. giop 10 = 0x00000400
	unsigned long activelow;	// GPIO bit reset to turn on LED
	unsigned long maskfail;
	unsigned long activelowfail;
} kled_info_t, *pkled_info_t;

pkled_info_t kled_get_object(BOARD_LED_NAME led);

static pkled_info_t g_phys_leds = NULL;
static pkled_info_t g_virt_leds[kLedEnd];

#define	GPIO_DIR	0xFFFE0084
#define	GPIO_DATA	0xFFFE008C

#define GPIO_NUM_TO_MASK(X)     (((X) < 32)? (1 << (X)) : 0)

//#define DEBUG_LED

#ifndef FALSE
#define FALSE           0
#endif /* FALSE */
#ifndef TRUE
#define TRUE            1
#endif /* TRUE */

/* to set the gpio mapping of buttons */
kled_pair_map_t kled_pair_mapping[] =
{
	[0] = {
		.name = kLedPower,
		.initstate = kLedStateSlowBlinkContinues,
		.mask = 0
	},
	[1] = {
		.name = kLedInternetRed,
		.mask = 1
	},
	[2] = {
		.name = kLedInternetGreen,
		.mask = 2
	},
	[3] = {
		.name = kLedVoip,
		.mask = 3
	},
	[4] = {
		.name = kLedLan,
		.mask = 4
	},
	[5] = {
		.name = kLedWireless,
		.mask = 5
	},
	[6] = {
		.name = kLedUpgrade,
		.mask = 6
	},
	[7] = {
		.name = kLedEnd
	}
};

/* to operate gpio to set on led */
void
kled_hw_set_on(BOARD_LED_NAME led)
{
	pkled_info_t pled = kled_get_object(led);

//	printk("kernel debugging - kled_hw_set_on() - led = %d, mask = %d.\n", led, pled->mask);

	if ((pled) && GPIO_NUM_TO_MASK(pled->mask))
	{
        	kSysLedCtrl(pled->mask, kLedStateOn);
	}
	return;
}

/* to operate gpio to set off led */
void
kled_hw_set_off(BOARD_LED_NAME led)
{
	pkled_info_t pled = kled_get_object(led);

	//printk("kernel debugging - kled_hw_set_off() - led = %d, mask = %d.\n", led, pled->mask);

	if ((pled) && GPIO_NUM_TO_MASK(pled->mask))
	{
        	kSysLedCtrl(pled->mask, kLedStateOff);
	}
	return;
}

/* to operate gpio to set on fail led */
void
kled_hw_set_onfail(BOARD_LED_NAME led)
{
	pkled_info_t pled = kled_get_object(led);

	if ((pled) && GPIO_NUM_TO_MASK(pled->maskfail))

	{
        	kSysLedCtrl(kLedInternetRed,kLedStateOn);
	}
}

/* to operate gpio to set off fail led */
void
kled_hw_set_offfail(BOARD_LED_NAME led)
{
	pkled_info_t pled = kled_get_object(led);

	if ((pled)  && GPIO_NUM_TO_MASK(pled->maskfail))
 
	{
        	kSysLedCtrl(kLedInternetRed, kLedStateOff);
	}
}

pkled_info_t
kled_get_object(BOARD_LED_NAME led)
{
	// mapping from virtual led to physical led
	if ((unsigned int)led < kLedEnd)
		return g_virt_leds[(unsigned int)led];
	else
		return NULL;
}

// Initialize a structure that contains information about a physical board LED
// control.  The board LED may contain more than one GPIO pin to control a
// normal condition (green) or a failure condition (red).
static void
_kled_init_info(pkled_info_t pled, pkled_pair_map_t pmap_pair)
{
	pled->mask = pmap_pair->mask;
	pled->activelow = pmap_pair->activelow;
	pled->maskfail = pmap_pair->maskfail;
	pled->activelowfail = pmap_pair->activelowfail;

	kled_init_state(pmap_pair->name, pmap_pair->initstate);
}

void __init
_kled_init_mapping(pkled_pair_map_t mapping)
{
	pkled_pair_map_t p1, p2;
	pkled_info_t pled;
	//int btimer = FALSE;
	int bused = FALSE;
	int led_count = 0;

	memset(g_virt_leds, 0x00, sizeof(g_virt_leds));

	// check for multiple led names and multiple led gpio pins that share the same physical leds on board
	for (p1 = mapping; p1->name != kLedEnd; p1++) {
		bused = FALSE;
		for (p2 = mapping; p2 != p1; p2++) {
			if (((p1->mask) && (p1->mask == p2->mask)) || ((p1->maskfail) && (p1->maskfail == p2->maskfail))) {
				bused = TRUE;
				break;
			}
		}
		if (FALSE == bused) led_count++;
	}

#ifdef DEBUG_LED
	printk("_kled_init_mapping() - %d led has been found.\n", led_count);
#endif /* DEBUG_LED */

	// allocate memory accoding to the exact number of on-board leds
	g_phys_leds = (pkled_info_t)kmalloc(led_count*sizeof(kled_info_t), GFP_KERNEL);
	if (NULL == g_phys_leds) {
		printk("LED memory allocation error\n");
		return;
	}
	memset(g_phys_leds, 0x00, (led_count * sizeof(kled_info_t)));

	// build led list
	pled = g_phys_leds;
	for (p1 = mapping; p1->name != kLedEnd; p1++) {
		if ((unsigned int)p1->name > kLedEnd) continue;

		bused = FALSE;
		for (p2 = mapping; p2 != p1; p2++) {
			if (((p1->mask) && (p1->mask == p2->mask)) || ((p1->maskfail) && (p1->maskfail == p2->maskfail))) {
				bused = TRUE;
				break;
			}
		}

		if (FALSE == bused) {
			_kled_init_info(pled, p1);
			g_virt_leds[(int)p1->name] = pled;
			// set an initial state as requested
			kled_set_state(p1->name, p1->initstate, 0, 0);
			pled++;
		} else {
			pkled_info_t pled2;

			for (pled2 = g_phys_leds; pled2 != pled; pled2++) {
				if (((pled2->mask) && (pled2->mask == p1->mask)) || ((pled2->maskfail) && (pled2->maskfail == p1->maskfail))) {
					if (p1->mask) {
						pled2->mask = p1->mask;
						pled2->activelow = p1->activelow;
					}
					if (p1->maskfail)
					{
						pled2->maskfail = p1->maskfail;
						pled2->activelowfail = p1->activelowfail;
					}
					g_virt_leds[(int)p1->name] = pled2;
					break;
				}
			}
		}
	}
#ifdef DEBUG_LED
	int i;
	for (i = 0; i < led_count; i++)
		printk("initLed: led[%d]: mask=0x%04x, activelow=%d\n", i, (g_phys_leds+i)->mask, (g_phys_leds+i)->activelow);
#endif /* DEBUG_LED */
}


#define MPI_CTRL_PCMCIA_CTRL1	0xFFFE1054
#define CARDBUS_ENABLE          0x00008000
#define PCMCIA_GPIO_ENABLE      0x00002000
/* to init gpio settings for led in kernel mode */
void
kled_hw_init(void)
{
	_kled_init_mapping(kled_pair_mapping);

	// TODO: put some line to init the GPIO here
	// if the two bits of pcmcia_control_1 are set, some GPIO will not work as normal IO.
#ifdef CONFIG_THOMSON_LIVEBOX_MINI
	*((volatile u32 *)MPI_CTRL_PCMCIA_CTRL1) &= ~(CARDBUS_ENABLE | PCMCIA_GPIO_ENABLE);
#endif /* CONFIG_THOMSON_LIVEBOX_MINI */
}

