/* vi: set sw=4 ts=4: */
/*
 *
 * mdev - Mini udev for busybox
 *
 * Copyright 2005 Rob Landley <rob@landley.net>
 * Copyright 2005 Frank Sorenson <frank@tuxrocks.com>
 *
 * Licensed under GPL version 2, see file LICENSE in this tarball for details.
 *
 * SoftAtHome:
 * - Added support for config files in /etc/mdev.d
 * - Added support for matching sysfs attribute:
 *   regex,ATTR{filename}=value
 */

#include "libbb.h"
#include "xregex.h"

struct globals {
	int root_major, root_minor;
};

struct device_action_info {
	const char *device_name;
        int delete;
        int major;
        int minor;
        int mode;
#if ENABLE_FEATURE_MDEV_CONF
        struct bb_uidgid_t ugid;
#endif
#if ENABLE_FEATURE_MDEV_EXEC
        char *command;
#endif
#if ENABLE_FEATURE_MDEV_RENAME
        char *alias;
        char aliaslink;
#endif
};

#define G (*(struct globals*)&bb_common_bufsiz1)
#define root_major (G.root_major)
#define root_minor (G.root_minor)

/* Prevent infinite loops in /sys symlinks */
#define MAX_SYSFS_DEPTH 3

/* We use additional 64+ bytes in make_device() */
#define SCRATCH_SIZE 80

#if ENABLE_FEATURE_MDEV_RENAME
/* Builds an alias path.
 * This function potentionally reallocates the alias parameter.
 */
static char *build_alias(char *alias, const char *device_name)
{
	char *dest;

	/* ">bar/": rename to bar/device_name */
	/* ">bar[/]baz": rename to bar[/]baz */
	dest = strrchr(alias, '/');
	if (dest) { /* ">bar/[baz]" ? */
		*dest = '\0'; /* mkdir bar */
		bb_make_directory(alias, 0755, FILEUTILS_RECUR);
		*dest = '/';
		if (dest[1] == '\0') { /* ">bar/" => ">bar/device_name" */
			dest = alias;
			alias = concat_path_file(alias, device_name);
			free(dest);
		}
	}

	return alias;
}
#endif

/* match a ATTR{file}=value attribute */
static int match_sysfs(struct device_action_info *da UNUSED_PARAM, char *attribute)
{
        char *c, *cb, *eq, *value, *file = attribute + 5;
        char dir[PATH_MAX + 1] = { 0, };
        char path[PATH_MAX + 1] = { 0, };
        char content[1025] = { 0, };
        char *parent;
        FILE *fd;
        int r;

        if ((cb = strchr(file, '}')) == NULL)
                return 0;

        *cb = '\0';
        if ((eq = strchr(cb + 1, '=')) == NULL) 
                return 0;

        value = eq + 1;

        /* check that the content of the file is matching the value */
        snprintf(dir, PATH_MAX, "/sys%s", getenv("DEVPATH"));

        do {
                snprintf(path, PATH_MAX, "%s/%s", dir, file);
                fd = fopen(path, "r");
                if (fd != NULL)
                        break;

                parent = dirname(dir);
                memmove(dir, parent, strlen(parent) + 1);
        } while (strncmp(dir, "/sys", PATH_MAX) != 0);

        if (fd == NULL)
                return 0;

        if ((r = fread(content, sizeof(char), 1024, fd)) < 0)
                return 0;
        
        /* strip trailing whitespaces */
        for (c = content + r - 1; c >= content && (isblank(*c) || iscntrl(*c)); c--) {
                *c = '\0';
        }
        for (c = value + r - 1; c >= value && (isblank(*c) || iscntrl(*c)); c--) {
                *c = '\0';
        }

        return (strcmp(value, content) == 0);
}

/* File callback for /etc/mdev.d/ traversal */
static int FAST_FUNC config_file_action(const char *filename,
                                        struct device_action_info *da)
{
	parser_t *parser;
	char *tokens[5];
        int found = FALSE;

        parser = config_open2(filename, fopen_for_read);


	/* If we have config file, look up user settings */
	while (config_read(parser, tokens, 4, 3, "# \t", PARSE_NORMAL)) {
		regmatch_t off[1 + 9*ENABLE_FEATURE_MDEV_RENAME_REGEXP];
		char *val;
		/* Fields: regex[,SYSFS{file}=value,...] uid:gid mode [alias] [cmd] */

		/* 1st field: @<numeric maj,min>... */
		if (tokens[0][0] == '@') {
			/* @major,minor[-last] */
			/* (useful when name is ambiguous:
			 * "/sys/class/usb/lp0" and
			 * "/sys/class/printer/lp0") */
			int cmaj, cmin0, cmin1, sc;
			if (da->major < 0)
				continue; /* no dev, no match */
			sc = sscanf(tokens[0], "@%u,%u-%u", &cmaj, &cmin0, &cmin1);
			if (sc < 1 || da->major != cmaj
			 || (sc == 2 && da->minor != cmin0)
			 || (sc == 3 && (da->minor < cmin0 || da->minor > cmin1))
			) {
				continue; /* no match */
			}
		} else { /* ... or regex to match device name */
			regex_t match;
			int result;
                        char *comma;
                        int attrmatch = 1;

                        /* check for , */
                        if ((comma = strchr(tokens[0], ',')) != NULL) {
                                *comma = '\0';
                        }

			/* Is this it? */
			xregcomp(&match, tokens[0], REG_EXTENDED);
			result = regexec(&match, da->device_name, ARRAY_SIZE(off), off, 0);
			regfree(&match);

			//bb_error_msg("matches:");
			//for (int i = 0; i < ARRAY_SIZE(off); i++) {
			//	if (off[i].rm_so < 0) continue;
			//	bb_error_msg("match %d: '%.*s'\n", i,
			//		(int)(off[i].rm_eo - off[i].rm_so),
			//		device_name + off[i].rm_so);
			//}

			/* If not this device, skip rest of line */
			/* (regexec returns whole pattern as "range" 0) */
			if (result || off[0].rm_so
			 || ((int)off[0].rm_eo != (int)strlen(da->device_name))
			) {
                                continue;
			}

                        /* if we have a , then try to match the attributes */
                        while (comma) {
                                char *attribute = comma + 1;
                                if ((comma = strchr(attribute, ',')) != NULL) 
                                        *comma = '\0';
                                if ((attribute = strstr(attribute, "ATTR{")) != NULL) {
                                        /* check if the attribute matches */
                                        if (!match_sysfs(da, attribute)) {
                                                attrmatch = 0;
                                                break;
                                        }
                                } else {
                                        /* we do not know about that attribute, assume not matching, continue */
                                        attrmatch = 0;
                                        break;
                                }
                                        
                                    
                        }
                        if (attrmatch == 0)
                                continue;

                        
		}

		/* This line matches: stop parsing the file
		 * after parsing the rest of fields */
                found = TRUE;

		/* 2nd field: uid:gid - device ownership */
		parse_chown_usergroup_or_die(&da->ugid, tokens[1]);

		/* 3rd field: mode - device permissions */
		da->mode = strtoul(tokens[2], NULL, 8);

		val = tokens[3];
		/* 4th field (opt): >alias */
#if ENABLE_FEATURE_MDEV_RENAME
		if (!val)
			break;
		da->aliaslink = *val;
		if (da->aliaslink == '>' || da->aliaslink == '=') {
			char *s;
#if ENABLE_FEATURE_MDEV_RENAME_REGEXP
			char *p;
			unsigned i, n;
#endif
			char *a = val;
			s = strchrnul(val, ' ');
			val = (s[0] && s[1]) ? s+1 : NULL;
			s[0] = '\0';
#if ENABLE_FEATURE_MDEV_RENAME_REGEXP
			/* substitute %1..9 with off[1..9], if any */
			n = 0;
			s = a;
			while (*s)
				if (*s++ == '%')
					n++;

			p = da->alias = xzalloc(strlen(a) + n * strlen(da->device_name));
			s = a + 1;
			while (*s) {
				*p = *s;
				if ('%' == *s) {
					i = (s[1] - '0');
					if (i <= 9 && off[i].rm_so >= 0) {
						n = off[i].rm_eo - off[i].rm_so;
						strncpy(p, da->device_name + off[i].rm_so, n);
						p += n - 1;
						s++;
					}
				}
				p++;
				s++;
			}
#else
			da->alias = xstrdup(a + 1);
#endif
		}
#endif /* ENABLE_FEATURE_MDEV_RENAME */

#if ENABLE_FEATURE_MDEV_EXEC
		/* The rest (opt): command to run */
		if (!val)
			break;
		{
			const char *s = "@$*";
			const char *s2 = strchr(s, *val);

			if (!s2)
				bb_error_msg_and_die("bad line %u", parser->lineno);

			/* Correlate the position in the "@$*" with the delete
			 * step so that we get the proper behavior:
			 * @cmd: run on create
			 * $cmd: run on delete
			 * *cmd: run on both
			 */
			if ((s2 - s + 1) /*1/2/3*/ & /*1/2*/ (1 + da->delete)) {
				da->command = xstrdup(val + 1);
			}
		}
#endif
		/* end of field parsing */
		break; /* we found matching line, stop */
	} /* end of "while line is read from /etc/mdev.conf" */
        
	config_close(parser);

        return found;
}

/* mknod in /dev based on a path like "/sys/block/hda/hda1" */
/* NB: "mdev -s" may call us many times, do not leak memory/fds! */
static void make_device(char *path, int delete)
{
	int type, len;
        struct device_action_info da;
	char *dev_maj_min = path + strlen(path);

        memset(&da, 0, sizeof(struct device_action_info));
        da.mode = 0660;
        da.delete = delete;

	/* Force the configuration file settings exactly. */
	umask(0);

	/* Try to read major/minor string.  Note that the kernel puts \n after
	 * the data, so we don't need to worry about null terminating the string
	 * because sscanf() will stop at the first nondigit, which \n is.
	 * We also depend on path having writeable space after it.
	 */
	da.major = -1;
	if (!delete) {
		strcpy(dev_maj_min, "/dev");
		len = open_read_close(path, dev_maj_min + 1, 64);
		*dev_maj_min++ = '\0';
		if (len < 1) {
			if (!ENABLE_FEATURE_MDEV_EXEC)
				return;
			/* no "dev" file, so just try to run script */
			*dev_maj_min = '\0';
		} else if (sscanf(dev_maj_min, "%u:%u", &da.major, &da.minor) != 2) {
			da.major = -1;
		}
	}


	/* Determine device name, type, major and minor */
	da.device_name = bb_basename(path);
	/* http://kernel.org/doc/pending/hotplug.txt says that only
	 * "/sys/block/..." is for block devices. "/sys/bus" etc is not.
	 * But since 2.6.25 block devices are also in /sys/class/block.
	 * We use strstr("/block/") to forestall future surprises. */
	type = S_IFCHR;
	if (strstr(path, "/block/"))
		type = S_IFBLK;

#if ENABLE_FEATURE_MDEV_CONF
        /* first check in /etc/mdev.conf */
        if (!config_file_action("/etc/mdev.conf", &da)) {
                struct dirent **namelist;
                int n, i = 0;
                n = scandir("/etc/mdev.d", &namelist, NULL, alphasort);
                for (i = 0; i < n; i++) {
                        char cpath[PATH_MAX + 1] = { 0, };
                        if (*(namelist[i]->d_name) == '.')
                                continue;

                        snprintf(cpath, PATH_MAX, "/etc/mdev.d/%s", namelist[i]->d_name);
                        free(namelist[i]);
                        if (config_file_action(cpath, &da) == 1)
                                break;
                }
                if (n >= 0)
                  free(namelist);
        }

#endif

	if (!delete && da.major >= 0) {

		if (ENABLE_FEATURE_MDEV_RENAME)
			unlink(da.device_name);

		if (mknod(da.device_name, da.mode | type, makedev(da.major, da.minor)) && errno != EEXIST)
			bb_perror_msg_and_die("mknod %s", da.device_name);

		if (da.major == root_major && da.minor == root_minor)
			symlink(da.device_name, "root");

#if ENABLE_FEATURE_MDEV_CONF
		chown(da.device_name, da.ugid.uid, da.ugid.gid);

#if ENABLE_FEATURE_MDEV_RENAME
		if (da.alias) {
			da.alias = build_alias(da.alias, da.device_name);

			/* move the device, and optionally
			 * make a symlink to moved device node */
			if (rename(da.device_name, da.alias) == 0 && da.aliaslink == '>')
				symlink(da.alias, da.device_name);

			free(da.alias);
		}
#endif
#endif
	}

#if ENABLE_FEATURE_MDEV_EXEC
	if (da.command) {
		/* setenv will leak memory, use putenv/unsetenv/free */
		char *s = xasprintf("MDEV=%s", da.device_name);
		putenv(s);
		if (system(da.command) == -1)
			bb_perror_msg_and_die("can't run '%s'", da.command);
		s[4] = '\0';
		unsetenv(s);
                /* this triggers a problem in valgrind and we are going to exit soon anyway */
		//free(s);
		free(da.command);
	}
#endif

	if (delete) {
		unlink(da.device_name);
		/* At creation time, device might have been moved
		 * and a symlink might have been created. Undo that. */
#if ENABLE_FEATURE_MDEV_RENAME
		if (da.alias) {
			da.alias = build_alias(da.alias, da.device_name);
			unlink(da.alias);
			free(da.alias);
		}
#endif
	}
}

/* File callback for /sys/ traversal */
static int FAST_FUNC fileAction(const char *fileName,
		struct stat *statbuf UNUSED_PARAM,
		void *userData,
		int depth UNUSED_PARAM)
{
	size_t len = strlen(fileName) - 4; /* can't underflow */
	char *scratch = userData;
        int envset = 0;
        char *devpath = NULL;

	/* len check is for paranoid reasons */
	if (strcmp(fileName + len, "/dev") != 0 || len >= PATH_MAX)
		return FALSE;

        /* if DEVPATH is not set, set it */
        if (!getenv("DEVPATH")) {
          envset = 1;
          devpath = xstrndup(fileName + 4, len - 4);
          setenv("DEVPATH", devpath, 0);
        }
	strcpy(scratch, fileName);
	scratch[len] = '\0';
	make_device(scratch, 0);
        if (envset) {
          unsetenv("DEVPATH");
          free(devpath);
        }

	return TRUE;
}

/* Directory callback for /sys/ traversal */
static int FAST_FUNC dirAction(const char *fileName UNUSED_PARAM,
		struct stat *statbuf UNUSED_PARAM,
		void *userData UNUSED_PARAM,
		int depth)
{
	return (depth >= MAX_SYSFS_DEPTH ? SKIP : TRUE);
}

/* For the full gory details, see linux/Documentation/firmware_class/README
 *
 * Firmware loading works like this:
 * - kernel sets FIRMWARE env var
 * - userspace checks /lib/firmware/$FIRMWARE
 * - userspace waits for /sys/$DEVPATH/loading to appear
 * - userspace writes "1" to /sys/$DEVPATH/loading
 * - userspace copies /lib/firmware/$FIRMWARE into /sys/$DEVPATH/data
 * - userspace writes "0" (worked) or "-1" (failed) to /sys/$DEVPATH/loading
 * - kernel loads firmware into device
 */
static void load_firmware(const char *const firmware, const char *const sysfs_path)
{
	int cnt;
	int firmware_fd, loading_fd, data_fd;

	/* check for /lib/firmware/$FIRMWARE */
	xchdir("/lib/firmware");
	firmware_fd = xopen(firmware, O_RDONLY);

	/* in case we goto out ... */
	data_fd = -1;

	/* check for /sys/$DEVPATH/loading ... give 30 seconds to appear */
	xchdir(sysfs_path);
	for (cnt = 0; cnt < 30; ++cnt) {
		loading_fd = open("loading", O_WRONLY);
		if (loading_fd != -1)
			goto loading;
		sleep(1);
	}
	goto out;

 loading:
	/* tell kernel we're loading by `echo 1 > /sys/$DEVPATH/loading` */
	if (full_write(loading_fd, "1", 1) != 1)
		goto out;

	/* load firmware by `cat /lib/firmware/$FIRMWARE > /sys/$DEVPATH/data */
	data_fd = open("data", O_WRONLY);
	if (data_fd == -1)
		goto out;
	cnt = bb_copyfd_eof(firmware_fd, data_fd);

	/* tell kernel result by `echo [0|-1] > /sys/$DEVPATH/loading` */
	if (cnt > 0)
		full_write(loading_fd, "0", 1);
	else
		full_write(loading_fd, "-1", 2);

 out:
	if (ENABLE_FEATURE_CLEAN_UP) {
		close(firmware_fd);
		close(loading_fd);
		close(data_fd);
	}
}

int mdev_main(int argc, char **argv) MAIN_EXTERNALLY_VISIBLE;
int mdev_main(int argc UNUSED_PARAM, char **argv)
{
	RESERVE_CONFIG_BUFFER(temp, PATH_MAX + SCRATCH_SIZE);

	/* We can be called as hotplug helper */
	/* Kernel cannot provide suitable stdio fds for us, do it ourself */
#if 1
	bb_sanitize_stdio();
#else
	/* Debug code */
	/* Replace LOGFILE by other file or device name if you need */
#define LOGFILE "/dev/console"
	/* Just making sure fd 0 is not closed,
	 * we don't really intend to read from it */
	xmove_fd(xopen("/", O_RDONLY), STDIN_FILENO);
	xmove_fd(xopen(LOGFILE, O_WRONLY|O_APPEND), STDOUT_FILENO);
	xmove_fd(xopen(LOGFILE, O_WRONLY|O_APPEND), STDERR_FILENO);
#endif

	xchdir("/dev");

	if (argv[1] && !strcmp(argv[1], "-s")) {
		/* Scan:
		 * mdev -s
		 */
		struct stat st;

		xstat("/", &st);
		root_major = major(st.st_dev);
		root_minor = minor(st.st_dev);

		/* ACTION_FOLLOWLINKS is needed since in newer kernels
		 * /sys/block/loop* (for example) are symlinks to dirs,
		 * not real directories.
		 * (kernel's CONFIG_SYSFS_DEPRECATED makes them real dirs,
		 * but we can't enforce that on users) */
		recursive_action("/sys/block",
			ACTION_RECURSE | ACTION_FOLLOWLINKS,
			fileAction, dirAction, temp, 0);
		recursive_action("/sys/class",
			ACTION_RECURSE | ACTION_FOLLOWLINKS,
			fileAction, dirAction, temp, 0);
	} else {
		char *seq;
		char *action;
		char *env_path;
		char seqbuf[sizeof(int)*3 + 2];
		int seqlen = seqlen; /* for compiler */

		/* Hotplug:
		 * env ACTION=... DEVPATH=... [SEQNUM=...] mdev
		 * ACTION can be "add" or "remove"
		 * DEVPATH is like "/block/sda" or "/class/input/mice"
		 */
		action = getenv("ACTION");
		env_path = getenv("DEVPATH");
		if (!action || !env_path)
			bb_show_usage();

		seq = getenv("SEQNUM");
		if (seq) {
			int timeout = 2000 / 32;
			do {
				seqlen = open_read_close("mdev.seq", seqbuf, sizeof(seqbuf-1));
				if (seqlen < 0)
					break;
				seqbuf[seqlen] = '\0';
				if (seqbuf[0] == '\n' /* seed file? */
				 || strcmp(seq, seqbuf) == 0 /* correct idx? */
				) {
					break;
				}
				usleep(32*1000);
			} while (--timeout);
		}

		snprintf(temp, PATH_MAX, "/sys%s", env_path);
		if (!strcmp(action, "remove"))
			make_device(temp, 1);
		else if (!strcmp(action, "add")) {
			make_device(temp, 0);

			if (ENABLE_FEATURE_MDEV_LOAD_FIRMWARE) {
				char *fw = getenv("FIRMWARE");
				if (fw)
					load_firmware(fw, temp);
			}
		}

		if (seq && seqlen >= 0) {
			xopen_xwrite_close("mdev.seq", utoa(xatou(seq) + 1));
		}
	}

	if (ENABLE_FEATURE_CLEAN_UP)
		RELEASE_CONFIG_BUFFER(temp);

	return 0;
}
